extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <assert.h>
#include <math.h>
}

#define CRSPDIR "/Users/ivo/research/Data/crsp/"

#include "wuni.hh"

/****************************************************************
 * Note: A stock that has many days but ceases to trade for a year can
 * continue on and produce a sigma with even one day after its
 * interim break data...because it had observations (very
 * downweighted) from long ago, it does not suffer the 3*(5-day WEEK)
 * minimum.
 ****************************************************************/

static inline double winsrel( double delta, double ri, double rm ) {
  double b1= (1.0+delta)*rm;
  double b2= (1.0-delta)*rm;
  if (b2 < b1) { double b3= b2; b2= b1; b1= b3; }

  // never cut off more below -0.02 and +0.02
  b1-= 0.02;
  b2+= 0.02;

  if (ri < b1) return b1;
  if (ri > b2) return b2;
  return ri;
}


int gregdt(int yyyymmdd) {
  int y= yyyymmdd / 10000;
  int m= (yyyymmdd / 100) % 100;
  int d= yyyymmdd % 100;

  y+=8000;
  if(m<3) { y--; m+=12; }
  return (y*365) +(y/4) -(y/100) +(y/400) -1200820
    +(m*153+3)/5-92
    +d-1 - (2424881 -1 - 153) ; // 19251231 is 0
}


/****************************************************************/
int main(int argc, char *argv[]) {
  double delta= (argc<=1) ? 999.0 : atof(argv[1]); // 999 means do not truncate
  double agedecay= (argc<=2) ? 0.0 : atof(argv[2]);  // 0 means do not decay

  fprintf(stderr, "## running %s: delta=%g agedecay=%g.  Freely Distributable.\n  Ivo Welch, \"Simpler Better Market Betas,\" SSRN 3371240\n\n", argv[0], delta, agedecay);

  assert( delta > 1.0 ); // or else the 1-delta definition in winsrel is not sensible

  char line[300];

  FILE *f= popen("gzcat " CRSPDIR "crspdaily-clean.csv.gz | cut -d, -f1,2,3,5,6,16", "r");
  assert( fgets(line, 300, f) != NULL );

  assert( strcmp( line, "permno,caldt,yyyymmdd,rf,vwretd,ret\n" ) == 0 );

  WUNI *wu=wuni_new();
  int linecount=0; int sigmaout=0;
  int lastpermno=10000, lastyyyymm=198601, lastyyyymmdd=19860108, lastdi=0;
  double curweight=1.0;

  printf("permno, gregdt , yyyymmdd , sd.d%g.a%g\n", delta, agedecay);
  while (!feof(f)) {
    ++linecount;

    // if ((linecount % 100000) == 0) fprintf(stderr, "[%d]\n", sigmaout);

    fgets(line, 200, f);
    if (strstr(line, ",,")!=NULL) continue;
    if (strstr(line, ",C,")!=NULL) continue;
    if (strstr(line, ",B,")!=NULL) continue;
    if (strstr(line, "B,")!=NULL) continue;

    int permno, caldt, yyyymmdd;  double rf, ret, vwretd;

    int numread=sscanf(line, "%d,%d,%d,%lf,%lf,%lf", &permno, &caldt, &yyyymmdd, &rf, &vwretd, &ret);
    // double xvwretd= vwretd-rf; double xret= ret-rf;
    double xvwretd= vwretd; double xret= ret;

    if (numread<5) {
      fprintf(stderr, "bswrets.csv.gz:%d: %d entries: %s\n", linecount, numread, line);
    }
    if (feof(f)) break;
    if (numread==5) { // missing return
      continue;
    }

    assert( numread==6 );
    assert( (yyyymmdd > 19200000) && (yyyymmdd < 22000000) );
    assert( xret > (-2.0) );
    assert( permno >= lastpermno );
    int di= gregdt( yyyymmdd );
    if (permno == lastpermno ) assert( di > lastdi );

    int yyyymm= yyyymmdd/100;

    // at least one week!!
    const int WEEK=5;
    if (((yyyymm != lastyyyymm) || (permno != lastpermno)) && (wu->n > (3*WEEK))) {
      // trailing day sigma to be printed
      double sigma= sqrt( wuni_var(wu) );
      sigma= winsrel( delta, sigma, 1.0 );  // some pathological cases
      printf("%d , %5d , %d ,%10.7f\n", lastpermno, lastdi, lastyyyymmdd, sigma);  // could add fregr->n
      ++sigmaout;
    }
    // double sigma= sqrt( wuni_var(wu) );
    // sigma= winsrel( delta, sigma, 1.0 );  // some pathological cases
    // fprintf(fdebug, " 2, %d , %5d , %d , s=%9.7f\n", lastpermno, lastdi, lastyyyymmdd, sigma);  // could add fregr->n

    if (permno != lastpermno) {
      wuni_clear(wu);
      curweight=1.0;
    }

    lastpermno=permno; 
    lastyyyymmdd=yyyymmdd;
    lastyyyymm= yyyymm;

    for (int i=0; i<(di-lastdi); ++i) curweight*= (1.0+agedecay/365); // a few weird cases with intermissions
    lastdi= di;

    double sxret= (delta < 998) ? winsrel( delta, xret, xvwretd ) : xret;
    wuni_include( wu, sxret, curweight );

    // if ((permno==10000)&&(yyyymm<=198607)) fprintf(stderr, "[in%06d:%d:%d] sxri=%f,xrm=%f [w=%f]\n", linecount, permno, lastyyyymmdd, sxret, xvwretd, curweight );
    // if (freg->wghtn >= 100.0)) {curweight/= 100.0;freg->rescale( 1.0/100.0 );}

  }
  pclose(f);

  fprintf(stderr, "DONE #delta=%g #agedecay=%g --> sigmas=%d\n", delta, agedecay, sigmaout);
  //  fprintf(stderr, "\nrun mkwrds2.R now\n");
  return 0;
}
